<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

/**
 * Service for handling Access Policy database content.
 *
 * This actually doesn't do a lot. This is primarily for easing the transition
 * to the base field storage. This will eventually be refactored out once all
 * storage has been settled.
 *
 * @deprecated in access_policy:1.0.0-beta1 and is removed from access_policy:1.0.0-beta2.
 *   There is no replacement.
 *
 * @package Drupal\access_policy
 */
class BaseFieldAccessPolicyStorage implements EntityAccessPolicyStorageInterface {

  /**
   * The database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * Constructs a ContentAccessPolicyManager object.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\access_policy\AccessPolicyInformation $policy_information
   *   Access policy information.
   */
  public function __construct(Connection $connection, EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $policy_information) {
    $this->connection = $connection;
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $policy_information;
  }

  /**
   * {@inheritdoc}
   */
  public function isUsed(AccessPolicyInterface $access_policy) {
    $entity_type = $access_policy->getTargetEntityTypeId();
    $table = $this->getTable($entity_type);
    $query = $this->connection->select($table);
    $result = $query->fields($table, ['entity_id'])
      ->condition('access_policy_target_id', $access_policy->id())
      ->range(0, 1)
      ->execute();

    $exists = $result->fetchField();

    return !empty($exists);
  }

  /**
   * {@inheritdoc}
   */
  public function getUsage(AccessPolicyInterface $access_policy) {
    $entity_type = $access_policy->getTargetEntityTypeId();
    $table = $this->getTable($entity_type);
    $query = $this->connection->select($table);
    $result = $query->fields($table, ['entity_id'])
      ->condition('access_policy_target_id', $access_policy->id())
      ->execute();

    return $result->fetchField();
  }

  /**
   * {@inheritdoc}
   */
  public function read(EntityInterface $entity) {
    $table = $this->getTable($entity->getEntityTypeId());
    $query = $this->connection->select($table);
    $query->fields($table, [
      'access_policy_target_id',
    ]);
    $query->condition('entity_id', $entity->id());
    $result = $query->execute();
    return $result->fetchCol();
  }

  /**
   * {@inheritdoc}
   */
  public function isNull(EntityInterface $entity) {
    $table = $this->getTable($entity->getEntityTypeId());
    $query = $this->connection->select($table);
    $query->condition('entity_id', $entity->id());
    $query->addExpression('COUNT(*)');
    $count = $query->execute()->fetchField();
    if ($count == 0) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function write(EntityInterface $entity, $access_policy) {
    // Handled by core.
  }

  /**
   * {@inheritdoc}
   */
  public function delete(EntityInterface $entity) {
    // Handled by core.
  }

  /**
   * {@inheritdoc}
   */
  public function getTable($entity_type) {
    return $entity_type . '__access_policy';
  }

  /**
   * {@inheritdoc}
   */
  public function getColumnName($key) {
    $column_map = [
      'entity_id' => 'entity_id',
      'value' => 'access_policy_target_id',
    ];

    if (isset($column_map[$key])) {
      return $column_map[$key];
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty($entity_type) {
    $table = $this->getTable($entity_type);
    $count = $this->connection->select($table)
      ->countQuery()
      ->execute()
      ->fetchField();
    if ($count == 0) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Install the storage table if it doesn't exist.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function installSchema(EntityTypeInterface $entity_type) {
    // Handled by core.
  }

  /**
   * Uninstall the storage table.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function uninstallSchema(EntityTypeInterface $entity_type) {
    // Handled by core.
  }

  /**
   * Get the storage schema.
   */
  public static function getSchema() {
    // Handled by core.
  }

}
