<?php

namespace Drupal\access_policy\Controller;

use Drupal\access_policy\AccessPolicySelectionInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\access_policy\AccessPolicyInformation;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Controller for entity Access tab.
 *
 * @package Drupal\access_policy\Controller
 */
class EntityAccessPolicy extends ControllerBase {

  /**
   * The current route.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInfo;

  /**
   * The access policy selection service.
   *
   * @var \Drupal\access_policy\AccessPolicySelectionInterface
   */
  protected $selection;

  /**
   * Constructs a new NodeAccessPolicyinstance.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route_match service.
   * @param \Drupal\access_policy\AccessPolicyInformation $access_policy_information
   *   The access policy information service.
   * @param \Drupal\access_policy\AccessPolicySelectionInterface $selection
   *   The access policy selection service.
   */
  public function __construct(RouteMatchInterface $route_match, AccessPolicyInformation $access_policy_information, AccessPolicySelectionInterface $selection) {
    $this->routeMatch = $route_match;
    $this->accessPolicyInfo = $access_policy_information;
    $this->selection = $selection;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('current_route_match'),
      $container->get('access_policy.information'),
      $container->get('access_policy.selection'),
    );
  }

  /**
   * Get the current entity.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The current entity.
   */
  protected function getEntity() {
    $entity_types = $this->accessPolicyInfo->getAllEnabledEntityTypes();
    foreach ($entity_types as $entity_type) {
      $entity = $this->routeMatch->getParameter($entity_type->id());
      if (!empty($entity)) {
        if (is_numeric($entity)) {
          return $this->loadEntity($entity_type->id(), $entity);
        }

        return $entity;
      }
    }

    return FALSE;
  }

  /**
   * Return the full entity object.
   *
   * In some cases the entity route might not be properly upcasted. Such as
   * with other contrib modules that are using a similar route path. This
   * makes sure that the entity is properly loaded and doesn't cause an error.
   *
   * @param string $entity_type_id
   *   The entity type id.
   * @param int $entity_id
   *   The entity id.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The current entity.
   */
  protected function loadEntity($entity_type_id, $entity_id) {
    $entity = $this->entityTypeManager()->getStorage($entity_type_id)->load($entity_id);
    if (!empty($entity)) {
      return $entity;
    }

    return FALSE;
  }

  /**
   * Access callback node Access tab.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The access result.
   */
  public function access() {
    $entity = $this->getEntity();

    // Check to see if this is page is enabled.
    if (!$this->selection->isSelectionPageEnabled($entity->getEntityTypeId())) {
      return AccessResult::forbidden();
    }

    // Check to see if this bundle supports access policy. If it doesn't then
    // do not show the Access tab.
    if (!empty($entity)) {
      if (!$this->accessPolicyInfo->isAccessControlledEntity($entity)) {
        return AccessResult::forbidden();
      }

      return AccessResult::allowed()->addCacheableDependency($entity);
    }

    return AccessResult::forbidden();
  }

}
