<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

/**
 * Service for handling Access Policy database content.
 *
 * @deprecated in access_policy:1.0.0-beta1 and is removed from access_policy:1.0.0-beta2.
 *   There is no replacement.
 *
 * @package Drupal\access_policy
 */
class EntityAccessPolicyStorage implements EntityAccessPolicyStorageInterface {

  /**
   * The database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * Constructs a ContentAccessPolicyManager object.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\access_policy\AccessPolicyInformation $policy_information
   *   Access policy information.
   */
  public function __construct(Connection $connection, EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $policy_information) {
    $this->connection = $connection;
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $policy_information;
  }

  /**
   * {@inheritdoc}
   */
  public function isUsed(AccessPolicyInterface $access_policy) {
    $entity_type = $access_policy->getTargetEntityTypeId();
    $table = $this->getTable($entity_type);
    $query = $this->connection->select($table);
    $result = $query->fields($table, ['entity_id'])
      ->condition('access_policy', $access_policy->id())
      ->range(0, 1)
      ->execute();

    $exists = $result->fetchField();

    return !empty($exists);
  }

  /**
   * {@inheritdoc}
   */
  public function getUsage(AccessPolicyInterface $access_policy) {
    $entity_type = $access_policy->getTargetEntityTypeId();
    $table = $this->getTable($entity_type);
    $query = $this->connection->select($table);
    $result = $query->fields($table, ['entity_id'])
      ->condition('access_policy', $access_policy->id())
      ->execute();

    return $result->fetchField();
  }

  /**
   * {@inheritdoc}
   */
  public function read(EntityInterface $entity) {
    $table = $this->getTable($entity->getEntityTypeId());
    $query = $this->connection->select($table);
    $query->fields($table, [
      'access_policy',
    ]);
    $query->condition('entity_id', $entity->id());
    $result = $query->execute();
    return $result->fetchCol();
  }

  /**
   * {@inheritdoc}
   */
  public function isNull(EntityInterface $entity) {
    $table = $this->getTable($entity->getEntityTypeId());
    $query = $this->connection->select($table);
    $query->condition('entity_id', $entity->id());
    $query->addExpression('COUNT(*)');
    $count = $query->execute()->fetchField();
    if ($count == 0) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function write(EntityInterface $entity, $access_policy) {

    if (empty($access_policy)) {
      $this->delete($entity);
      return;
    }

    if (!is_array($access_policy)) {
      $access_policy = [$access_policy];
    }

    $this->verifyPolicyExists($access_policy);

    $transaction = $this->connection->startTransaction();
    try {
      $this->delete($entity);

      // Confirm that the policies are supported by this entity type.
      $policies = $this->entityTypeManager->getStorage('access_policy')->loadMultiple($access_policy);
      foreach ($policies as $policy) {
        $enabled = $this->accessPolicyInformation->isEnabledForEntitiesOfEntityType($policy, $entity->getEntityType());
        if (!$enabled) {
          throw new \Exception(sprintf('The "%1$s" Access Policy is not enabled for "%2$s"', $policy->id(), $entity->getEntityTypeId()));
        }
      }

      // Also confirm that if there are more than one policy then they are all
      // part of the same set.
      if (count($policies) > 1) {
        $this->verifyMatchingSelectionSets($policies);
      }

      $table = $this->getTable($entity->getEntityTypeId());
      $query = $this->connection->insert($table)->fields([
        'bundle',
        'entity_id',
        'access_policy',
      ]);

      foreach ($access_policy as $policy) {
        $values = [
          'bundle' => $entity->bundle(),
          'entity_id' => $entity->id(),
          'access_policy' => $policy,
        ];
        $query->values($values);
      }
      $query->execute();
    }
    catch (\Exception $e) {
      $transaction->rollBack();
    }

    unset($transaction);
  }

  /**
   * Verify that the access policies exist.
   *
   * @param array $access_policies
   *   The array of access policy ids.
   *
   * @throws \Exception
   */
  protected function verifyPolicyExists(array $access_policies) {
    foreach ($access_policies as $policy_id) {
      if (!empty($policy_id)) {
        $policy = $this->entityTypeManager->getStorage('access_policy')->load($policy_id);
        if (empty($policy)) {
          throw new \Exception(sprintf('The "%1$s" access policy does not exist.', $policy_id));
        }
      }
    }
  }

  /**
   * Verify that the access policies can be assigned together.
   *
   * @param array $access_policies
   *   Array of access policies.
   *
   * @throws \Exception
   */
  protected function verifyMatchingSelectionSets(array $access_policies) {
    $selection_sets = [];
    foreach ($access_policies as $policy) {
      $set = $policy->getSelectionSet();
      $selection_sets = array_merge($set, $selection_sets);
    }
    $count_values = array_count_values($selection_sets);
    $allow = FALSE;
    foreach ($count_values as $count) {
      // If the counts match for any of the sets then it is allowed to be
      // assigned.
      if ($count == count($access_policies)) {
        $allow = TRUE;
      }
    }
    if (!$allow) {
      throw new \Exception(sprintf('The selection sets do not match.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function delete(EntityInterface $entity) {
    $table = $this->getTable($entity->getEntityTypeId());
    $this->connection->delete($table)
      ->condition('entity_id', $entity->id())
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function getTable($entity_type) {
    return $entity_type . '_access_policy';
  }

  /**
   * {@inheritdoc}
   */
  public function getColumnName($key) {
    $column_map = [
      'entity_id' => 'entity_id',
      'value' => 'access_policy',
    ];

    if (isset($column_map[$key])) {
      return $column_map[$key];
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty($entity_type) {
    $table = $this->getTable($entity_type);
    $count = $this->connection->select($table)
      ->countQuery()
      ->execute()
      ->fetchField();
    if ($count == 0) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Install the storage table if it doesn't exist.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function installSchema(EntityTypeInterface $entity_type) {
    if (!$this->accessPolicyInformation->isSupportedEntityType($entity_type)) {
      throw new \Exception(sprintf('The "%1$s" entity type is not supported by access policy.', $entity_type->id()));
    }

    $table = $this->getTable($entity_type->id());
    $schema = $this->connection->schema();
    if (!$schema->tableExists($table)) {
      $table_schema = self::getSchema();
      $schema->createTable($table, $table_schema);
    }
  }

  /**
   * Uninstall the storage table.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function uninstallSchema(EntityTypeInterface $entity_type) {
    $table = $this->getTable($entity_type->id());
    $schema = $this->connection->schema();
    if (!$schema->tableExists($table)) {
      $schema->dropTable($table);
    }
  }

  /**
   * Get the storage schema.
   *
   * @return array
   *   The entity access policy storage schema.
   */
  public static function getSchema() {
    return [
      'description' => 'Stores entity access policy data.',
      'fields' => [
        'bundle' => [
          'description' => 'The bundle.',
          'type' => 'varchar_ascii',
          'length' => 128,
          'not null' => TRUE,
          'default' => '',
        ],
        'entity_id' => [
          'description' => 'The entity id.',
          'type' => 'int',
          'unsigned' => TRUE,
          'not null' => TRUE,
          'default' => 0,
        ],
        'access_policy' => [
          'description' => 'The access policy value.',
          'type' => 'varchar_ascii',
          'length' => 128,
          'not null' => TRUE,
          'default' => '',
        ],
      ],
      'primary key' => ['entity_id', 'access_policy'],
    ];
  }

}
