<?php

namespace Drupal\access_policy\Form;

use Drupal\access_policy\Plugin\AccessPolicyHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;

/**
 * The form for adding access rules to an access policy.
 *
 * This is actually a pseudo form, it's being loaded as part of
 * AccessRuleMultistepForm. Not in the route.
 *
 * @internal
 */
class AccessPolicyAddRuleForm extends AccessRuleFormBase {

  /**
   * Field map static cache.
   *
   * @var array
   */
  protected $fieldMap = [];

  /**
   * The access policy entity object.
   *
   * @var \Drupal\access_policy\Entity\AccessPolicyInterface
   */
  protected $accessPolicy;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_add_rule_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, AccessPolicyInterface $access_policy = NULL, $type = NULL) {

    $this->accessPolicy = $access_policy;
    $this->handlerType = $type;

    $form['#attributes']['class'][] = 'handler-modal-form';

    $form['filter'] = [
      '#title' => $this->t('Search'),
      '#type' => 'textfield',
      '#ajax' => [
        'event' => 'change',
        'keypress' => TRUE,
        'callback' => [$this, 'ajaxFilterCallback'],
      ],
    ];

    // Show access rule plugins (if available).
    $definitions = $this->getApplicableRules();

    // Sort the access rules by label.
    $definitions = $this->sortAccessRules($definitions);

    // Filter the access rules if applicable.
    if (!empty($form_state->getValue('filter'))) {
      $definitions = $this->filterAccessRules($definitions, $form_state->getValue('filter'));
    }

    $options = [];
    foreach ($definitions as $id => $access_rule) {
      $options[$id] = [
        'title' => $access_rule->getDefinition()->getLabel(),
        'description' => $this->getDescription($access_rule),
      ];
    }

    $form['access_rules'] = [
      '#type' => 'tableselect',
      '#title' => $this->t('Access rule'),
      '#js_select' => FALSE,
      '#header' => [
        'title' => $this->t('Title'),
        'description' => $this->t('Description'),
      ],
      '#element_validate' => [[static::class, 'addRuleValidate']],
      '#options' => $options,
      '#empty' => $this->t('No handlers found.'),
    ];

    return $form;
  }

  /**
   * Get the description of the access policy handler.
   *
   * @param \Drupal\access_policy\Plugin\AccessPolicyHandlerInterface $handler
   *   The access policy handler plugin.
   *
   * @return string
   *   The description.
   */
  protected function getDescription(AccessPolicyHandlerInterface $handler) {
    $description = $handler->getDefinition()->getDescription();
    if (!empty($description)) {
      return $description;
    }

    // Get the "Appears in" text if no description is provided.
    if ($handler->getDefinition()->getFieldName()) {
      $field = $handler->getDefinition()->getFieldName();

      if (empty($this->fieldMap)) {
        $fieldManager = \Drupal::service('entity_field.manager');
        $this->fieldMap = $fieldManager->getFieldMap();
      }

      $entity_type = $handler->getDefinition()->getEntityType();
      if (!empty($this->fieldMap[$entity_type][$field]['bundles'])) {
        $bundles = $this->fieldMap[$entity_type][$field]['bundles'];
        $labels = $this->getBundleLabels($entity_type, $bundles);
        $bundles = implode(', ', $labels);
        return $this->t('Appears in: @bundles', ['@bundles' => $bundles]);
      }
    }

    return '';
  }

  /**
   * Get the bundle labels.
   *
   * @param string $entity_type
   *   The entity type.
   * @param array $bundles
   *   The array of bundle ids.
   *
   * @return array
   *   Array of bundle labels.
   */
  protected function getBundleLabels($entity_type, array $bundles) {
    $bundleInfo = \Drupal::service('entity_type.bundle.info');

    $labels = [];
    $bundle_info = $bundleInfo->getBundleInfo($entity_type);
    foreach ($bundles as $bundle) {
      $labels[] = $bundle_info[$bundle]['label'];
    }

    return $labels;
  }

  /**
   * Validate the access rule selection.
   *
   * @param array $form
   *   The current form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public static function addRuleValidate(array $form, FormStateInterface $form_state) {
    $access_rules = $form_state->getValue('access_rules');
    $selected_rules = array_filter($access_rules);
    if (empty($selected_rules)) {
      $form_state->setErrorByName('access_rules', 'Please select a rule.');
    }
    if (count($selected_rules) > 1) {
      $form_state->setErrorByName('access_rules', 'Only one rule can be added at a time.');
    }
  }

  /**
   * Ajax callback to reload the form filtered access rules.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   An ajax response object.
   */
  public function ajaxFilterCallback(array $form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $response->addCommand(new ReplaceCommand('.handler-modal-form', $form));
    return $response;
  }

  /**
   * Filter access rules.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   * @param string $search
   *   The search string.
   *
   * @return array
   *   Array of filtered definitions.
   */
  public function filterAccessRules(array $access_rules, $search) {
    return array_filter($access_rules, function ($rule) use ($search) {
      if (stripos($rule->getDefinition()->getLabel(), $search) !== FALSE) {
        return TRUE;
      }
      return FALSE;
    });
  }

  /**
   * Sort the access rules by label.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   *
   * @return array
   *   Array of sorted access rule plugins.
   */
  protected function sortAccessRules(array $access_rules) {
    $key_value = [];
    foreach ($access_rules as $id => $rule) {
      $key_value[$id] = $rule->getDefinition()->getLabel();
    }
    natsort($key_value);

    $sorted = [];
    foreach ($key_value as $id => $label) {
      $sorted[$id] = $access_rules[$id];
    }

    return $sorted;
  }

  /**
   * Returns all applicable access rules for this policy.
   *
   * @return array
   *   Array of applicable access rules.
   */
  public function getApplicableRules() {
    $supported_groups = [
      'global',
      'user',
      $this->accessPolicy->getTargetEntityTypeId(),
    ];

    $rules = [];
    foreach ($supported_groups as $group) {
      $group_rules = $this->getHandlerManager()->getHandlersByGroup($group);
      $rules = array_merge($rules, $group_rules);
    }
    return $rules;
  }

}
