<?php

namespace Drupal\access_policy\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * Class AccessPolicyCreateForm.
 *
 * The AccessPolicy create form.
 *
 * @package Drupal\access_policy\Form
 */
class AccessPolicyCreateForm extends EntityForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $entity = $this->entity;

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $entity->label(),
      '#description' => $this->t("The human-readable name of this access policy. This name must be unique."),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $entity->id(),
      '#machine_name' => [
        'exists' => '\Drupal\access_policy\Entity\AccessPolicy::load',
      ],
      '#disabled' => !$entity->isNew(),
    ];

    $form['target_entity_type_id'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#default_value' => 'node',
      '#options' => $this->getContentEntityTypeOptions(),
      '#description' => $this->t('Choose the entity type that this policy is for.'),
    ];

    $form['type'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Access policy type'),
    ];
    $types = $this->getAccessPolicyTypes();
    foreach ($types as $definition) {
      $form['type'][$definition['id']] = [
        '#title' => $definition['label'],
        '#parents' => ['type'],
        '#type' => 'radio',
        '#default_value' => $this->entity->getType(),
        '#return_value' => $definition['id'],
        '#description' => $definition['description'],
        '#required' => TRUE,
      ];
    }

    return $form;
  }

  /**
   * Get all the supported content entity types.
   *
   * @return array
   *   Array of content entity types options, keyed by entity type id.
   */
  private function getContentEntityTypeOptions() {
    $entity_types = $this->entityTypeManager->getDefinitions();
    $content_entity_types = array_filter($entity_types, function (EntityTypeInterface $entity_type) {
      return \Drupal::service('access_policy.information')->isSupportedEntityType($entity_type);
    });
    $options = [];
    foreach ($content_entity_types as $entity_type) {
      $options[$entity_type->id()] = $entity_type->getLabel();
    }

    return $options;
  }

  /**
   * Get all the access policy types.
   *
   * @return array
   *   Array of access policy types definitions.
   */
  private function getAccessPolicyTypes() {
    return \Drupal::service('plugin.manager.access_policy_type')->getDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $entity = $this->entity;

    $status = $entity->save();

    switch ($status) {
      case SAVED_NEW:
        $this->messenger()->addMessage($this->t('Created the %label Access Policy.', [
          '%label' => $entity->label(),
        ]));

        // Warn users to configure permissions when creating a new policy.
        $type = $entity->getTypePlugin();
        if ($type->supportsPermissions()) {
          $message = $this->t('This access policy uses permissions. <a href = ":perms">Configure permissions</a> to grant access.', [
            ':perms' => Url::fromRoute('entity.access_policy.permissions_form', ['access_policy' => $entity->id()])->toString(),
          ]);
          $this->messenger()->addWarning($message);
        }
        break;
    }
    $form_state->setRedirectUrl($entity->toUrl('edit-form'));
  }

}
