<?php

namespace Drupal\access_policy\Form;

use Drupal\access_policy\AccessPolicyHandlerManager;
use Drupal\access_policy\EntityTypeSettingsInterface;
use Drupal\access_policy\Http403ResponsePluginManager;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\access_policy\AccessPolicyInformation;

/**
 * Class AccessPolicyEditForm.
 *
 * The AccessPolicy edit form.
 *
 * @package Drupal\access_policy\Form
 */
class AccessPolicySelectionForm extends EntityForm {

  /**
   * The AccessPolicyHandlerManager service.
   *
   * @var \Drupal\access_policy\AccessPolicyHandlerManager
   */
  protected $accessRulePluginManager;

  /**
   * The access denied behavior manager.
   *
   * @var \Drupal\access_policy\Http403ResponsePluginManager
   */
  protected $http403ResponsePluginManager;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * The access policy entity type settings.
   *
   * @var \Drupal\access_policy\EntityTypeSettingsInterface
   */
  protected $entityTypeSettings;

  /**
   * Constructs a AccessPolicyEdiForm object.
   *
   * @param \Drupal\access_policy\AccessPolicyHandlerManager $access_rule_manager
   *   The AccessPolicyHandlerManager service.
   * @param \Drupal\access_policy\Http403ResponsePluginManager $http_403_manager
   *   The access denied behavior manager.
   * @param \Drupal\access_policy\AccessPolicyInformation $access_policy_information
   *   The access policy information service.
   * @param \Drupal\access_policy\EntityTypeSettingsInterface $settings
   *   The access policy entity type settings.
   */
  public function __construct(AccessPolicyHandlerManager $access_rule_manager, Http403ResponsePluginManager $http_403_manager, AccessPolicyInformation $access_policy_information, EntityTypeSettingsInterface $settings) {
    $this->accessRulePluginManager = $access_rule_manager;
    $this->http403ResponsePluginManager = $http_403_manager;
    $this->accessPolicyInformation = $access_policy_information;
    $this->entityTypeSettings = $settings;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.access_policy.access_rule'),
      $container->get('plugin.manager.http_403_response'),
      $container->get('access_policy.information'),
      $container->get('access_policy.entity_type_settings')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $entity = $this->entity;

    $form['selection_rule_container'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Selection rules'),
    ];

    $form['selection_rule_container']['help'] = [
      '#type' => 'markup',
      '#markup' => $this->t('Selection rules allow you to define which policies are available based on different attributes. For example: original author, bundle, field value, etc.'),
    ];
    $form['selection_rule_container']['access_rules_table'] = $this->buildAccessRuleTable($entity);

    $form['selection_rule_container']['add_rule'] = [
      '#type' => 'link',
      '#title' => $this->t('Add selection rule'),
      '#url' => Url::fromRoute('access_policy.access_policy_add_rule_form', [
        'access_policy' => $entity->id(),
        'type' => 'selection_rule',
      ]),
      '#attributes' => $this->getModalAttributes(),
    ];

    $form['selection_rule_container']['selection_rule_operator'] = [
      '#type' => 'select',
      '#title' => $this->t('Validation operator'),
      '#description' => $this->t('How many rules must pass in order to assign this policy?'),
      '#default_value' => $entity->getSelectionRuleOperator(),
      '#options' => [
        'AND' => 'All',
        'OR' => 'Any',
      ],
    ];

    $form['selection_set_container'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Multiple policy support'),
    ];
    $form['selection_set_container']['help'] = [
      '#markup' => $this->t('By default, only one access policy can be assigned at a time. To use this policy with others, please add it to a selection set.'),
    ];
    if (empty($this->getSelectionSetOptions())) {
      $form['selection_set_container']['selection_set'] = [
        '#markup' => "<p>" . $this->t('No selection sets available. Add selection sets from <a href = ":link">:entity_type settings</a>.', [
          ':entity_type' => $entity->getTargetEntityType()->getLabel(),
          ':link' => Url::fromRoute('access_policy.entity_type_settings_form', ['target_entity_type_id' => $entity->getTargetEntityTypeId()])->toString(),
        ]) . "</p>",
      ];
    }
    else {
      $form['selection_set_container']['selection_set'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Selection set'),
        '#options' => $this->getSelectionSetOptions(),
        '#default_value' => $entity->getSelectionSet(),
      ];
    }

    return $form;
  }

  /**
   * Get the array of available selection sets.
   *
   * @return array
   *   Array of selection sets.
   */
  public function getSelectionSetOptions() {
    $settings = $this->entityTypeSettings->load($this->entity->getTargetEntityTypeId());
    $selection_sets = $settings->get('selection_sets') ?? [];
    $options = [];
    foreach ($selection_sets as $set) {
      $options[$set['id']] = $set['label'];
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $entity = $this->entity;

    $values = $form_state->getValues();

    $selection_set = [];
    if (!empty($values['selection_set'])) {
      $selection_set = array_keys(array_filter($values['selection_set']));
    }
    $entity->setSelectionSet($selection_set);

    $entity->save();

    $this->messenger()->addMessage($this->t('Saved the %label Access Policy.', [
      '%label' => $entity->label(),
    ]));
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    if (isset($actions['delete'])) {
      unset($actions['delete']);
    }

    return $actions;
  }

  /**
   * The access rule plugin manager.
   *
   * @return \Drupal\access_policy\AccessPolicyHandlerManager
   *   The access rule plugin manager.
   */
  protected function accessRuleManager() {
    return \Drupal::service('plugin.manager.access_policy.access_rule');
  }

  /**
   * Build the full access rule table.
   *
   * @return array
   *   The fully formed table of access rules.
   */
  public function buildAccessRuleTable(AccessPolicyInterface $policy) {
    $header = [
      'type' => $this->t('Items'),
      'operations' => $this->t('Operations'),
    ];
    $selected_rules_list = [
      '#type' => 'table',
      '#header' => $header,
      '#empty' => $this->t('No handlers selected.'),
      '#attributes' => [
        'id' => 'access-rules',
      ],
    ];
    $selected_rules = $this->getHandlerManager()->getHandlersFromPolicy($policy);
    foreach ($selected_rules as $id => $plugin) {
      $selected_rules_list[$id] = [
        '#attributes' => [
          'id' => 'rule--' . $id,
        ],
        'type' => [
          '#type' => 'markup',
          '#markup' => $plugin->adminLabel(),
        ],
        'operations' => [
          '#type' => 'operations',
          '#links' => $this->buildOperationsLinks($policy, $id),
        ],
      ];
    }

    return $selected_rules_list;
  }

  /**
   * Build the operations links.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $accessPolicy
   *   The access policy entity.
   * @param string $id
   *   The access rule id.
   *
   * @return array
   *   Array of operations links.
   */
  protected function buildOperationsLinks(AccessPolicyInterface $accessPolicy, $id) {
    $operations['configure'] = [
      'title' => $this->t('Configure'),
      'url' => Url::fromRoute('access_policy.access_policy_configure_rule_form', [
        'access_policy' => $accessPolicy->id(),
        'type' => 'selection_rule',
        'id' => $id,
      ]),
      'attributes' => $this->getModalAttributes(),
    ];

    // If the plugin is required then don't allow it to be removed.
    // For some reason it's not enforcing route access on the operation link.
    // We have to do it ourselves.
    $delete_url = Url::fromRoute('access_policy.access_policy_remove_rule_form', [
      'access_policy' => $accessPolicy->id(),
      'type' => 'selection_rule',
      'id' => $id,
    ]);
    if ($delete_url->access(\Drupal::currentUser())) {
      $operations['remove'] = [
        'title' => $this->t('Remove'),
        'url' => $delete_url,
        'attributes' => $this->getModalAttributes(),
      ];
    }

    return $operations;
  }

  /**
   * Get the modal attributes.
   *
   * @return array
   *   The modal attributes.
   */
  protected function getModalAttributes() {
    return [
      'class' => ['use-ajax'],
      'data-dialog-type' => 'modal',
      'data-dialog-options' => Json::encode([
        'width' => '75%',
      ]),
    ];
  }

  /**
   * Get the access policy handler manager.
   *
   * @return \Drupal\access_policy\AccessPolicyHandlerManager
   *   The selection rule plugin manager.
   */
  protected function getHandlerManager() {
    return \Drupal::service('plugin.manager.access_policy.selection_rule');
  }

}
