<?php

namespace Drupal\access_policy\Form;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\access_policy\Entity\AccessPolicyInterface;

/**
 * Provides a trait for handling the access rule table.
 */
trait AccessRuleTableTrait {

  /**
   * Build the full access rule table.
   *
   * @return array
   *   The fully formed table of access rules.
   */
  public function buildAccessRuleTable(AccessPolicyInterface $policy) {
    $header = [
      'type' => $this->t('Items'),
      'operations' => $this->t('Operations'),
    ];
    $selected_rules_list = [
      '#type' => 'table',
      '#header' => $header,
      '#empty' => $this->t('No handlers selected.'),
      '#attributes' => [
        'id' => 'access-rules',
      ],
    ];
    $selected_handlers = $this->getHandlerManager()->getHandlersFromPolicy($policy);
    foreach ($selected_handlers as $id => $plugin) {
      $selected_rules_list[$id] = [
        '#attributes' => [
          'id' => 'rule--' . $id,
        ],
        'type' => [
          '#type' => 'markup',
          '#markup' => $plugin->adminLabel(),
        ],
        'operations' => [
          '#type' => 'operations',
          '#links' => $this->buildOperationsLinks($policy, $id),
        ],
      ];
    }

    return $selected_rules_list;
  }

  /**
   * Build the operations links.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $accessPolicy
   *   The access policy entity.
   * @param string $id
   *   The access rule id.
   *
   * @return array
   *   Array of operations links.
   */
  protected function buildOperationsLinks(AccessPolicyInterface $accessPolicy, $id) {
    $operations['configure'] = [
      'title' => $this->t('Configure'),
      'url' => Url::fromRoute('access_policy.access_policy_configure_rule_form', [
        'access_policy' => $accessPolicy->id(),
        'type' => $this->handlerType,
        'id' => $id,
      ]),
      'attributes' => $this->getModalAttributes(),
    ];

    // If the plugin is required then don't allow it to be removed.
    // For some reason it's not enforcing route access on the operation link.
    // We have to do it ourselves.
    $delete_url = Url::fromRoute('access_policy.access_policy_remove_rule_form', [
      'access_policy' => $accessPolicy->id(),
      'type' => $this->handlerType,
      'id' => $id,
    ]);
    if ($delete_url->access(\Drupal::currentUser())) {
      $operations['remove'] = [
        'title' => $this->t('Remove'),
        'url' => $delete_url,
        'attributes' => $this->getModalAttributes(),
      ];
    }

    return $operations;
  }

  /**
   * Get the modal attributes.
   *
   * @return array
   *   The modal attributes.
   */
  protected function getModalAttributes() {
    return [
      'class' => ['use-ajax'],
      'data-dialog-type' => 'modal',
      'data-dialog-options' => Json::encode([
        'width' => '75%',
      ]),
    ];
  }

}
