<?php

namespace Drupal\access_policy\Plugin\AccessPolicyType;

/**
 * Access policy types.
 *
 * These plugins control how policies are validated. They control whether
 * permissions and/or rules should be validated for each operation (e.g view,
 * update and delete).
 *
 * This is useful in cases where you want to create a policy but a "view"
 * permission restriction might not make sense for it (like with Public
 * content).
 */
interface AccessPolicyTypeInterface {

  /**
   * Determine whether permissions are validated as part of this type.
   *
   * @param string $op
   *   The operation (e.g view, update, and delete).
   *
   * @return bool
   *   TRUE if permissions should be validated; FALSE otherwise.
   */
  public function shouldValidatePermission($op);

  /**
   * Determine whether rules are validated as part of this type.
   *
   * @param string $op
   *   The operation (e.g view, update and delete).
   *
   * @return bool
   *   TRUE if rules should be validated; FALSE otherwise.
   */
  public function shouldValidateRules($op);

  /**
   * Determine whether rule validation is enabled.
   *
   * All policy types can support both enabled and disabled access rule
   * validation. This allows us to toggle it at the plugin level.
   *
   * @param string $op
   *   The operation.
   *
   * @return bool
   *   TRUE if the rule validation is enabled; FALSE otherwise.
   */
  public function ruleValidationEnabled($op);

  /**
   * Return the supported operations.
   *
   * @return string[]
   *   Array of operations (e.g view, update, delete, assign).
   */
  public function getAllOperations();

  /**
   * Get all operations controlled by rules.
   *
   * Operations can be toggled per access rule so this returns an associative
   * array instead of indexed array.
   *
   * @return array
   *   Associative array of operations with flag indicating whether they are
   *   enabled or not. (e.g ['view' => TRUE, 'update' => FALSE, ...]).
   */
  public function getRuleOperations();

  /**
   * Get all operations controlled by permissions.
   *
   * @return array
   *   Array of operations.
   */
  public function getPermissionOperations();

  /**
   * Determine whether this type supports permission validation.
   *
   * @return bool
   *   TRUE if it's supported; FALSE otherwise.
   */
  public function supportsPermissions();

  /**
   * Determine whether this type supports rule validation.
   *
   * @return bool
   *   TRUE if it's supported; FALSE otherwise.
   */
  public function supportsRules();

  /**
   * Get the type access rule configuration.
   *
   * @return array
   *   Array of access rule configuration.
   */
  public function getTypeRules();

  /**
   * Determine whether this type has access rules defined.
   *
   * @return bool
   *   TRUE if type rules exist; FALSE otherwise.
   */
  public function hasTypeRules();

  /**
   * Get the selection rule configuration.
   *
   * @return array
   *   Array of selection rule configuration.
   */
  public function getTypeSelectionRules();

  /**
   * Determine whether this type has selection rules defined.
   *
   * @return bool
   *   TRUE if selection rules exist; FALSE otherwise.
   */
  public function hasTypeSelectionRules();

}
