<?php

namespace Drupal\access_policy\Plugin\SelectionStrategy;

use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Discretionary selection strategy.
 *
 * @SelectionStrategy(
 *   id = "discretionary",
 *   label = "Discretionary",
 *   description = @Translation("Users assign policies from an Access tab. Similar to traditional document-based access control."),
 *   weight = 10,
 * )
 */
class DiscretionarySelection extends SelectionStrategyBase {

  /**
   * {@inheritdoc}
   */
  public function defaultOptions() {
    return [
      'dynamic_assignment' => 'never',
      'enable_selection_page' => TRUE,
      'show_operations_link' => FALSE,
      'enable_policy_field' => TRUE,
      'allow_empty' => FALSE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultSettings() {
    return [
      'allow_empty' => TRUE,
      'default_policy' => 'empty',
      'show_operations_link' => FALSE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(EntityTypeInterface $entity_type) {
    return $entity_type->hasLinkTemplate('canonical');
  }

  /**
   * {@inheritdoc}
   */
  public function buildSettingsForm(array $form, FormStateInterface $form_state) {

    $form['show_operations_link'] = [
      '#title' => $this->t('Show operations link'),
      '#type' => 'checkbox',
      '#description' => $this->t('Show a link to the Access tab on listing pages.'),
      '#default_value' => $this->settings['show_operations_link'],
    ];

    $form['allow_empty'] = [
      '#title' => $this->t('Allow empty'),
      '#type' => 'checkbox',
      '#default_value' => $this->settings['allow_empty'],
      '#description' => $this->t('Allow authors to create and edit :entity_type without an access policy.', [':entity_type' => $this->entityType->getPluralLabel()]),
    ];

    $empty_allowed = [':input[name="strategy_settings[allow_empty]"]' => ['checked' => FALSE]];

    $form['default_policy'] = [
      '#title' => $this->t('Default policy'),
      '#type' => 'radios',
      '#options' => [
        'empty' => $this->t('Do not assign a policy'),
        'first_available' => $this->t('First available'),
      ],
      '#default_value' => $this->settings['default_policy'],
      '#states' => [
        'disabled' => $empty_allowed,
      ],
      '#description' => $this->t('Set the default policy when new :entity_type are created.', [':entity_type' => $this->entityType->getPluralLabel()]),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareOptions(array $options) {
    $options = parent::prepareOptions($options);

    if ($this->settings['default_policy'] == 'first_available') {
      $options['dynamic_assignment'] = 'on_create';
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function submitSettingsForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getErrors()) {
      $form_state->cleanValues();
      $values = $form_state->getValue('strategy_settings');
      $this->settings['show_operations_link'] = (bool) $values['show_operations_link'];
      $this->settings['allow_empty'] = (bool) $values['allow_empty'];

      if (!$this->settings['allow_empty']) {
        $this->settings['default_policy'] = 'first_available';
      }
      else {
        $this->settings['default_policy'] = $values['default_policy'];
      }
    }
  }

}
