<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Tests the creation and deletion of access policy entities.
 *
 * @group access_policy
 */
class AccessPolicyCreationTest extends AccessPolicyTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'datetime',
    'options',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Tests creating and deleting a new access policy with rules and bundles.
   */
  public function testCreateAndDeleteAccessPolicyWithRules() {
    $settings = $this->entityTypeSettings->load('node');
    $settings->set('selection_strategy', 'discretionary');
    $settings->save();

    $rid = $this->drupalCreateRole([
      'access administration pages',
      'administer access policy entities',
    ]);
    $web_user = $this->drupalCreateUser();
    $web_user->addRole($rid);
    $web_user->save();

    $this->drupalLogin($web_user);

    $this->drupalGet('admin/people/access-policies');
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/people/access-policies/add');

    $edit = [
      'label' => 'Foo',
      'id' => 'foo',
      'type' => 'group',
    ];
    $this->submitForm($edit, 'Save');

    // Confirm that permissions warning appears on new policies.
    $this->assertSession()->pageTextContains('This access policy uses permissions.');

    // Add a rule to the policy (cancel)
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/add');
    $this->assertSession()->statusCodeEquals(200);
    $edit = [
      'access_rules[node:title]' => TRUE,
    ];
    $this->submitForm($edit, 'Add rule');
    // Go to access rule configuration form and press cancel. Make sure it
    // does not save anything.
    $this->submitForm([], 'Cancel');
    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->pageTextNotContains('Field value: Title (= )');

    // Add a rule to the policy (submit).
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/add');
    $this->assertSession()->statusCodeEquals(200);
    $edit = [
      'access_rules[node:title]' => TRUE,
    ];
    $this->submitForm($edit, 'Add rule');
    $this->assertSession()->fieldValueEquals('query', TRUE);
    $this->assertSession()->fieldValueEquals('admin_label', '');
    $this->submitForm([], 'Save');

    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->pageTextContains('Field value: Title (= )');

    // Change settings of title access rule.
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/title');
    $this->assertSession()->statusCodeEquals(200);
    $edit = [
      'operator' => '=',
      'value' => 'Foo bar',
      'query' => FALSE,
    ];
    $this->submitForm($edit, 'Save');

    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->pageTextContains('Field value: Title (= Foo bar)');

    $this->drupalGet('/admin/people/access-policies/foo/access_rule/title');
    $this->assertSession()->fieldValueEquals('operator', '=');
    $this->assertSession()->fieldValueEquals('value', 'Foo bar');
    $this->assertSession()->fieldValueEquals('query', FALSE);

    // Change the admin label.
    $edit = [
      'admin_label' => 'Title field overridden.',
    ];
    $this->submitForm($edit, 'Save');
    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->pageTextContains('Title field overridden.');

    // Delete the is_own access rule.
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/title/remove');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Are you sure you want to remove this rule?');
    $this->submitForm([], 'Remove');
    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->pageTextNotContains('Authored by current user (modified)');

    // Change the other fields in the policy.
    $edit = [
      'label' => 'Bar',
      'description' => 'Lorem ipsum dolor.',
      'access_rule_operator' => 'OR',
    ];
    $this->submitForm($edit, 'Save');

    $this->clickLink('Configure');

    $this->assertSession()->fieldValueEquals('label', 'Bar');
    $this->assertSession()->fieldValueEquals('description', 'Lorem ipsum dolor.');
    $this->assertSession()->fieldValueEquals('access_rule_operator', 'OR');

    $this->clickLink('Delete');
    $this->assertSession()->pageTextContains('Are you sure you want to delete Bar?');
    $this->submitForm([], 'Delete');
    $this->assertSession()->pageTextContains('Deleted Bar Access Policy');

    $policy = $this->entityTypeManager->getStorage('access_policy')->load('foo');
    $this->assertEmpty($policy, 'The policy should have been deleted.');
  }

  /**
   * Tests setting creating a new access policy.
   */
  public function testDoNotPermitPolicyDeletionWhenUsed() {
    AccessPolicy::create([
      'id' => 'test_policy',
      'label' => 'Test policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ])->save();

    $node = $this->drupalCreateNode([
      'type' => 'page',
    ]);
    $this->assignAccessPolicy($node, 'test_policy');

    $web_user = $this->drupalCreateUser([
      'access administration pages',
      'administer access policy entities',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('admin/people/access-policies/test_policy/delete');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Are you sure you want to delete Test policy?');
    $this->assertSession()->pageTextContains('The access policy Test policy is being used, and cannot be deleted.');
    $this->assertSession()->linkNotExists('Delete');
  }

  /**
   * Tests creating a private access policy.
   */
  public function testCreatePrivateAccessPolicy() {

    $web_user = $this->drupalCreateUser([
      'access administration pages',
      'administer access policy entities',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('admin/people/access-policies');
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/people/access-policies/add');

    $edit = [
      'label' => 'Foo',
      'id' => 'foo',
      'type' => 'private',
    ];
    $this->submitForm($edit, 'Save');

    $this->assertSession()->pageTextContains('Authored by current user');

    // Confirm that permissions warning does not appear because private
    // policies do not support permissions.
    $this->assertSession()->pageTextNotContains('This access policy uses permissions.');

    // Make sure that the delete link is not visible nor accessible.
    $this->assertSession()->linkByHrefNotExists('admin/people/access-policies/foo/access_rule/is_own/remove');
    $this->drupalGet('admin/people/access-policies/foo/access_rule/is_own/remove');
    $this->assertSession()->statusCodeEquals(403);

    // Assert that the operations exists.
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/is_own');
    $this->assertSession()->statusCodeEquals(200);

    // This should not be using the default operations.
    $this->assertSession()->checkboxChecked('use_default_ops');
    $this->assertSession()->checkboxChecked('operations[view]');
    $this->assertSession()->checkboxChecked('operations[view all revisions]');
    $this->assertSession()->checkboxChecked('operations[update]');
    $this->assertSession()->checkboxChecked('operations[delete]');
    $this->assertSession()->checkboxChecked('operations[view unpublished]');

    // Add a rule to the policy and make sure that "assign access policy" is
    // disabled.
    $this->drupalGet('/admin/people/access-policies/foo/access_rule/add');
    $this->assertSession()->statusCodeEquals(200);
    $edit = [
      'access_rules[node:type]' => TRUE,
    ];
    $this->submitForm($edit, 'Add rule');

    // This should be using the default operations.
    $this->assertSession()->checkboxChecked('use_default_ops');
    $this->assertSession()->checkboxChecked('operations[view]');
    $this->assertSession()->checkboxChecked('operations[update]');
    $this->assertSession()->checkboxChecked('operations[delete]');
    $this->assertSession()->checkboxChecked('operations[view unpublished]');
    $this->assertSession()->checkboxChecked('operations[view all revisions]');

    $submit = [
      'operations[view]' => 1,
      'operations[update]' => 1,
      'operations[delete]' => 1,
      'operations[view unpublished]' => 1,
      'operations[view all revisions]' => 1,
    ];
    $this->submitForm($submit, 'Save');

    $this->drupalGet('/admin/people/access-policies/foo');
    $this->assertSession()->statusCodeEquals(200);

    // Verify the access policy operation storage.
    // If the operations matches the default policy type operations it should
    // not save it as part of policy settings. It's unnecessary noise.
    $policy = $this->entityTypeManager->getStorage('access_policy')->load('foo');
    $type_rule = $policy->getAccessRule('type');
    $empty = !isset($type_rule['settings']['operations']);
    $this->assertTrue($empty, 'The access rule should not store the operations');

  }

  /**
   * Tests creating a new public access policy.
   */
  public function testCreatePublicPolicy() {

    $web_user = $this->drupalCreateUser([
      'access administration pages',
      'administer access policy entities',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('admin/people/access-policies');
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/people/access-policies/add');

    $edit = [
      'label' => 'Foo',
      'id' => 'foo',
      'type' => 'public',
    ];
    $this->submitForm($edit, 'Save');

    $this->assertSession()->pageTextNotContains('Access Rules');
    $this->assertSession()->linkNotExists('Add Rule');
    $this->assertSession()->linkNotExists('Validation operator');
  }

  /**
   * Tests creating and viewing an access policy with broken access rules.
   */
  public function testCreatePolicyWithBrokenRule() {
    // This is an access policy with the old access rule data and a missing
    // plugin. Without a valid plugin it is effectively broken.
    AccessPolicy::create([
      'id' => 'broken_rules',
      'label' => 'Access policy with broken rules.',
      'access_rules' => [
        'field_value:node:field_text' => [
          'plugin_id' => 'field_value:node:field_text',
          'settings' => [
            'target' => 'value',
            'value' => 'My text',
            'operator' => '=',
            'query' => TRUE,
            'empty_behavior' => 'deny',
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
    ])->save();

    // This is an access policy with correct access rule data. This acts as a
    // control to compare against the broken_rules policy.
    $policy = AccessPolicy::create([
      'id' => 'working_rules',
      'label' => 'Access policy with working rules.',
      'target_entity_type_id' => 'node',
    ]);
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text', [
      'value' => 'My text',
      'operator' => '=',
      'query' => TRUE,
      'empty_behavior' => 'deny',
    ]);

    $policy->addAccessRule($access_rule);
    $policy->save();

    $rid = $this->drupalCreateRole([
      'access administration pages',
      'access content overview',
      'administer access policy entities',
      'view broken_rules content',
    ]);
    $web_user = $this->drupalCreateUser();
    $web_user->addRole($rid);
    $web_user->save();

    $this->drupalLogin($web_user);

    // Because the user does not have an entity reference this normally
    // would return 403, but because it's broken it should return 200.
    $node_broken = $this->drupalCreateNode([
      'title' => 'Node (broken)',
      'type' => 'page',
      'status' => 1,
    ]);
    $node_working = $this->drupalCreateNode([
      'title' => 'Node (working)',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'My text',
    ]);
    $this->assignAccessPolicy($node_broken, 'broken_rules');
    $this->assignAccessPolicy($node_working, 'working_rules');

    $this->drupalGet('node/' . $node_working->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('node/' . $node_broken->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce('Node (broken)');
    $this->assertSession()->pageTextNotContains('Node (working)');

    // Configure the policy.
    $this->drupalGet('admin/people/access-policies/broken_rules');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce('Broken/missing handler');

    // Configure the broken access rule.
    $this->clickLink('Configure');
    $this->assertSession()->pageTextContains('The handler for this access rule is broken or missing.');

    // Delete the broken access rule.
    $this->drupalGet('/admin/people/access-policies/broken_rules/access_rule/field_value:node:field_text/remove');
    $this->submitForm([], 'Remove');
    $this->drupalGet('/admin/people/access-policies/broken_rules');
    $this->assertSession()->pageTextNotContains('Broken/missing handler');

    // Add a rule to the policy. It should show the broken handler.
    $this->drupalGet('/admin/people/access-policies/broken_rules/access_rule/add');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Broken/missing handler');
  }

  /**
   * Tests creating and viewing an access policy invalid access rule data.
   */
  public function testCreatePolicyWithInvalidAccessRuleData() {
    // This is an access policy with the old access rule data. However, this has
    // a valid plugin id, therefore it should still work but show a warning.
    AccessPolicy::create([
      'id' => 'invalid_rules',
      'label' => 'Access policy with invalid rule data.',
      'access_rules' => [
        'field_text' => [
          'plugin_id' => 'entity_field_string',
          'settings' => [
            'target' => 'value',
            'value' => 'My text',
            'operator' => '=',
            'query' => TRUE,
            'empty_behavior' => 'deny',
            'admin_label' => '',
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
    ])->save();

    // This is an access policy with valid access rule data. This acts as a
    // control to compare against the invalid_rules policy.
    $policy = AccessPolicy::create([
      'id' => 'valid_rules',
      'label' => 'Access policy with valid rule data.',
      'target_entity_type_id' => 'node',
    ]);
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text', [
      'value' => 'My text',
      'operator' => '=',
      'query' => TRUE,
      'empty_behavior' => 'deny',
    ]);

    $policy->addAccessRule($access_rule);
    $policy->save();

    $rid = $this->drupalCreateRole([
      'access administration pages',
      'access content overview',
      'administer access policy entities',
      'view invalid_rules content',
      'view valid_rules content',
    ]);
    $web_user = $this->drupalCreateUser();
    $web_user->addRole($rid);
    $web_user->save();

    $this->drupalLogin($web_user);

    $node_invalid = $this->drupalCreateNode([
      'title' => 'Node (invalid)',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Not matching text',
    ]);
    $node_valid = $this->drupalCreateNode([
      'title' => 'Node (valid)',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Not matching text',
    ]);
    $this->assignAccessPolicy($node_invalid, 'invalid_rules');
    $this->assignAccessPolicy($node_valid, 'valid_rules');

    $this->drupalGet('node/' . $node_valid->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('node/' . $node_invalid->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextNotContains('Node (invalid)');
    $this->assertSession()->pageTextNotContains('Node (valid)');

    // Configure the policy.
    $this->drupalGet('admin/people/access-policies/invalid_rules');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce('Warning: invalid data detected');

    // It should be possible to configure it.
    $this->clickLink('Configure');
    $this->assertSession()->fieldExists('operations[view]');
    $this->assertSession()->fieldExists('admin_label');

    // Delete the invalid access rule.
    $this->drupalGet('/admin/people/access-policies/invalid_rules/access_rule/field_text/remove');
    $this->submitForm([], 'Remove');
    $this->drupalGet('/admin/people/access-policies/invalid_rules');
    $this->assertSession()->pageTextNotContains('Warning: invalid data detected');
  }

  /**
   * Confirm that immutable operator access rules do not show the operator.
   */
  public function testCreateAccessRuleWithImmutableOperator() {
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_user',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'user',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_user',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    $policy = AccessPolicy::create([
      'id' => 'basic_policy',
      'label' => 'Basic policy',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_user_user_reference');
    $policy->addAccessRule($access_rule);
    $policy->save();

    $rid = $this->drupalCreateRole([
      'access administration pages',
      'access content overview',
      'administer access policy entities',
    ]);
    $web_user = $this->drupalCreateUser();
    $web_user->addRole($rid);
    $web_user->save();

    $this->drupalLogin($web_user);

    $this->drupalGet('/admin/people/access-policies/basic_policy');
    $this->assertSession()->pageTextContains('field_user field has reference to current user');
    $this->clickLink('Configure');

    // Confirm that the operator field is not shown.
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldNotExists('operator');
  }

  /**
   * Tests the field access permission setting on selection rules.
   */
  public function testCreateSelectionRuleWithFieldAccessPermission() {
    $policy = AccessPolicy::create([
      'id' => 'basic_policy',
      'label' => 'Basic policy',
      'target_entity_type_id' => 'node',
      'weight' => -10,
    ]);
    $policy->save();

    $selection_rule = $this->selectionRuleManager->getHandler('node', 'field_text');
    $selection_rule->setSettings([
      'operator' => 'not empty',
    ]);
    $policy->addHandler('selection_rule', $selection_rule);
    $policy->save();

    $rid = $this->drupalCreateRole([
      'access content',
      'access administration pages',
      'access content overview',
      'edit any page content',
      'create page content',
      'administer access policy entities',
      'assign basic_policy access policy',
      'view basic_policy content',
      'edit basic_policy content',
    ]);
    $admin_user = $this->drupalCreateUser();
    $admin_user->addRole($rid);
    $admin_user->save();

    $this->drupalLogin($admin_user);

    $this->drupalGet('/admin/people/access-policies/basic_policy/selection_rule/field_text');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Field access');
    $this->assertSession()->fieldExists('field_access[type]');

    $edit = [
      'field_access[type]' => 'permission',
      'field_access[permission]' => 'assign basic_policy access policy',
    ];
    $this->submitForm($edit, 'Save');

    // Make sure it saved properly.
    $this->drupalGet('/admin/people/access-policies/basic_policy/selection');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Text (not empty)');

    $policy = $this->entityTypeManager->getStorage('access_policy')->load($policy->id());
    $rules = $policy->getHandlers('selection_rule');
    $rule_settings = $rules['field_text']['settings'];
    $this->assertEquals('permission', $rule_settings['field_access']['type']);
    $this->assertEquals('assign basic_policy access policy', $rule_settings['field_access']['permission']);

    // Assign the access policy to a node and edit it. Make sure that the field
    // is hidden.
    $node = $this->drupalCreateNode([
      'title' => 'Node',
      'type' => 'page',
      'status' => 1,
    ]);
    // Edit the node, confirm that they can see the field.
    $this->drupalGet('node/' . $node->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldExists('field_text[0][value]');

    // Create a new node. They should be able to see the field.
    $this->drupalGet('node/add/page');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldExists('field_text[0][value]');

    // This user can't assign the basic policy, therefore should not be able
    // to see the field.
    $web_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'edit any page content',
      'create page content',
    ]);

    $this->drupalLogin($web_user);

    // Edit the node, confirm that they can't see the field.
    $this->drupalGet('node/' . $node->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldNotExists('field_text[0][value]');

    // Create a new node. They should not be able to see the field.
    $this->drupalGet('node/add/page');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldNotExists('field_text[0][value]');

    // Create a new node that is assigned the Basic policy. Make sure that
    // It doesn't change the policy if another user without permission to
    // assign that policy edits it.
    $this->drupalLogin($admin_user);
    $node_2 = $this->drupalCreateNode([
      'title' => 'Node 2',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Foo bar',
    ]);
    $node_2 = $this->drupalGetNodeByTitle($node_2->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node_2);
    $policy = reset($policy);
    $this->assertEquals('basic_policy', $policy->id());

    // This user can edit Basic policy content node but can't assign the Basic
    // policy. Make sure that it doesn't reset the policy.
    $web_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'edit any page content',
      'create page content',
      'edit basic_policy content',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node_2->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldNotExists('field_text[0][value]');

    // Save the node, make sure that it doesn't change the policy.
    $edit = [
      'title[0][value]' => 'Node 2',
    ];
    $this->submitForm($edit, 'Save');

    $this->entityTypeManager->getStorage('node')->resetCache();
    $node = $this->drupalGetNodeByTitle('Node 2');
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $policy = reset($policy);
    $this->assertEquals('basic_policy', $policy->id(), 'This policy is incorrect');
  }

  /**
   * Tests viewing user profile with field observed by access rule.
   *
   * This tests multiple field types because some access rules use argument
   * settings to determine which field to observe.
   */
  public function testCreateAccessRuleUserFieldPermissions() {

    // Create a list field for user and node.
    FieldStorageConfig::create([
      'field_name' => 'field_list_string',
      'entity_type' => 'node',
      'type' => 'list_string',
      'cardinality' => 1,
      'settings' => [
        'allowed_values' => [
          'foo' => 'Foo',
          'bar' => 'Bar',
        ],
      ],
    ])->save();

    FieldConfig::create([
      'field_name' => 'field_list_string',
      'entity_type' => 'node',
      'label' => 'Test options list field',
      'bundle' => 'page',
    ])->save();

    FieldStorageConfig::create([
      'field_name' => 'field_list_string',
      'entity_type' => 'user',
      'type' => 'list_string',
      'cardinality' => 1,
      'settings' => [
        'allowed_values' => [
          'foo' => 'Foo',
          'bar' => 'Bar',
        ],
      ],
    ])->save();

    FieldConfig::create([
      'field_name' => 'field_list_string',
      'entity_type' => 'user',
      'label' => 'Test options list field',
      'bundle' => 'user',
    ])->save();

    FieldStorageConfig::create([
      'field_name' => 'field_boolean',
      'entity_type' => 'user',
      'type' => 'boolean',
    ])->save();

    FieldConfig::create([
      'field_name' => 'field_boolean',
      'entity_type' => 'user',
      'bundle' => 'user',
      'label' => "boolean",
      'required' => TRUE,
      'settings' => [
        'on_label' => 'On',
        'off_label' => 'Off',
      ],
    ])->save();

    // Add the new field to the edit form.
    EntityFormDisplay::create([
      'targetEntityType' => 'user',
      'bundle' => 'user',
      'mode' => 'default',
      'status' => TRUE,
    ])
      ->setComponent('field_boolean', [
        'type' => 'boolean_checkbox',
      ])
      ->setComponent('field_list_string', [
        'type' => 'options_select',
      ])
      ->save();

    $policy = AccessPolicy::create([
      'id' => 'basic_policy',
      'label' => 'Basic policy',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('user', 'field_boolean');
    $policy->addAccessRule($access_rule);

    $access_rule = $this->accessRuleManager->getHandler('node', 'match_field_list_string');
    $access_rule->setSettings([
      'operator' => 'in',
      'value' => ['field' => 'field_list_string'],
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $web_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer access policy entities',
    ]);

    $this->drupalLogin($web_user);

    $this->drupalGet('/user/' . $web_user->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldNotExists('field_list_string');
    $this->assertSession()->fieldNotExists('field_boolean');

    // Create a new user with the edit user information permission.
    $web_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer access policy entities',
      'edit basic_policy user information',
    ]);

    $this->drupalLogin($web_user);

    $this->drupalGet('/user/' . $web_user->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldExists('field_list_string');
    $this->assertSession()->fieldNotExists('field_boolean');
  }

}
