<?php

namespace Drupal\Tests\access_policy\Kernel;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\Core\Session\UserSession;
use Drupal\node\Entity\NodeType;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\Entity\Vocabulary;

/**
 * Tests Access Policy permissions.
 *
 * @group access_policy
 */
class AccessPolicyOperationsAccessTest extends AccessPolicyKernelTestBase {

  /**
   * The modules required for this test.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'taxonomy',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy']);
    $this->installEntitySchema('taxonomy_term');

    $vocabulary = Vocabulary::create([
      'name' => 'tags',
      'vid' => 'tags',
    ]);
    $vocabulary->save();

    $node_type = NodeType::create([
      'type' => 'page',
      'label' => 'Page',
    ]);
    $node_type->save();

    $this->contentAccessPolicyManager = $this->container->get('access_policy.content_policy_manager');
  }

  /**
   * Testing Group Access Policy operation access without rules.
   */
  public function testPublicAccessPolicyOperations() {
    $policy = AccessPolicy::create([
      'id' => 'public_policy',
      'label' => 'Public test policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'type' => 'public',
    ]);
    $policy->save();

    $node = $this->createNode(['type' => 'page']);
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Create a new revision, this is necessary for revert/delete revisions.
    $node->setNewRevision();
    $node->set('title', 'New revision');
    $node->save();

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'view page revisions',
      'revert page revisions',
      'delete page revisions',
      'edit public_policy content',
      'delete public_policy content',
      'view all public_policy content revisions',
      'set entity access policy',
      'assign public_policy access policy',
    ], 'public_policy');

    $account = new UserSession([
      'uid' => 2,
      'roles' => ['public_policy'],
    ]);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content Access page.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view the revision.');

    $result = $node_revision->access('delete revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to delete the revision.');

    $result = $node_revision->access('revert revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to revert the revision.');

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'view page revisions',
      'revert page revisions',
      'delete page revisions',
      'set entity access policy',
    ], 'authenticated');

    $account = new UserSession([
      'roles' => ['authenticated'],
    ]);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the public content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to assign the policy.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see the revision.');

    $result = $node_revision->access('revert revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to revert the revision.');

    $result = $node_revision->access('delete revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to delete the revision.');

  }

  /**
   * Testing Group Access Policy operation access without rules.
   */
  public function testGroupAccessPolicyOperationsWithoutRules() {
    $policy = AccessPolicy::create([
      'id' => 'group_policy',
      'label' => 'Group test policy without rules',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    $node = $this->createNode(['type' => 'page']);
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Create a new revision, this is necessary for revert/delete revisions.
    $node->setNewRevision();
    $node->set('title', 'New revision');
    $node->save();

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'view group_policy content',
      'edit group_policy content',
      'delete group_policy content',
      'view all group_policy content revisions',
      'set entity access policy',
      'assign group_policy access policy',
    ], 'group_policy');

    $account = new UserSession([
      'uid' => 2,
      'roles' => ['group_policy'],
    ]);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content Access page.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view the revision.');

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'set entity access policy',
    ], 'authenticated');

    $account = new UserSession([
      'roles' => ['authenticated'],
    ]);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to assign the policy.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see all revisions.');

    $result = $node->access('view revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see the revision.');

  }

  /**
   * Testing Group Access Policy operation access with rules.
   */
  public function testGroupAccessPolicyOperationsWithRules() {
    $policy = AccessPolicy::create([
      'id' => 'group_policy',
      'label' => 'Group test policy with rules',
      'target_entity_type_id' => 'node',
    ]);
    $access_rule = $this->accessRuleManager->getHandler('node', 'is_own');
    $policy->addAccessRule($access_rule);
    $policy->save();

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'view group_policy content',
      'edit group_policy content',
      'delete group_policy content',
      'view all group_policy content revisions',
      'set entity access policy',
      'assign group_policy access policy',
    ], 'group_policy');

    $account = $this->createUser();
    $account->addRole('group_policy');
    $account->save();

    $node = $this->createNode(['type' => 'page', 'uid' => $account->id()]);
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Create a new revision, this is necessary for revert/delete revisions.
    $node->setNewRevision();
    $node->set('title', 'New revision');
    $node->save();

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content Access page.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view the revision.');

    // This is a different user account that has the same group permissions
    // but the rule fails preventing them from getting access to everything
    // except assignment.
    $account = $this->createUser();
    $account->addRole('group_policy');
    $account->save();

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to assign the policy.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see all revisions.');

    $result = $node->access('view revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to see the revision.');

  }

  /**
   * Testing Private Access Policy operation access.
   */
  public function testPrivateAccessPolicyOperations() {
    $policy = AccessPolicy::create([
      'id' => 'private_policy',
      'label' => 'Private test policy with only rules.',
      'target_entity_type_id' => 'node',
      'type' => 'private',
    ]);
    $policy->save();

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'set entity access policy',
      'assign private_policy access policy',
    ], 'test_role');

    $account = $this->createUser([], 'user1', FALSE, ['uid' => 2]);
    $account->addRole('test_role');
    $account->save();

    $node = $this->createNode(['type' => 'page', 'uid' => $account->id()]);
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Create a new revision, this is necessary for revert/delete revisions.
    $node->setNewRevision();
    $node->set('title', 'New revision');
    $node->save();

    $this->setCurrentUser($account);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content Access page.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view the revision.');

    $account = $this->createUser([], 'user2', FALSE, ['uid' => 3]);
    $account->addRole('test_role');
    $account->save();
    $this->setCurrentUser($account);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user is not the original author and should not be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user is not the original author and should not able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user is not the original author and should not be able to delete the content.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user is not the original author and should not be able to see all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user is not the original author and should not be able to see the revision.');
  }

  /**
   * Testing Drupal default Access Policy operation access.
   *
   * This is mostly just testing drupal core access except for the assign
   * operation.
   */
  public function testDefaultAccessPolicyOperations() {
    $policy = AccessPolicy::create([
      'id' => 'default_policy',
      'label' => 'Default test policy.',
      'target_entity_type_id' => 'node',
      'type' => 'default',
    ]);
    $policy->save();

    $this->createRole([
      'access content',
      'edit any page content',
      'delete any page content',
      'view all revisions',
      'set entity access policy',
      'assign default_policy access policy',
    ], 'test_role');

    $account = $this->createUser([], 'user1', FALSE, ['uid' => 2]);
    $account->addRole('test_role');
    $account->save();

    $node = $this->createNode(['type' => 'page', 'uid' => $account->id()]);
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Create a new revision, this is necessary for revert/delete revisions.
    $node->setNewRevision();
    $node->set('title', 'New revision');
    $node->save();

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content Access page.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be be able to view the revision.');

    $account = $this->createUser([
      'access content',
    ]);

    $result = $node->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the content.');

    $result = $node->access('update', $account, TRUE);
    $this->assertTrue($result->isNeutral(), 'The user should not able to edit the content.');

    $result = $node->access('delete', $account, TRUE);
    $this->assertTrue($result->isNeutral(), 'The user should not be able to delete the content.');

    $result = $node->access('manage access', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to assign the policy.');

    $result = $node->access('view all revisions', $account, TRUE);
    $this->assertTrue($result->isNeutral(), 'The user should not be able to see all revisions.');

    $node_revision = \Drupal::service('entity_type.manager')->getStorage('node')->loadRevision(1);
    $result = $node_revision->access('view revision', $account, TRUE);
    $this->assertTrue($result->isNeutral(), 'The user should not be able to see the revision.');
  }

  /**
   * Test access policy group permissions with taxonmy terms.
   */
  public function testTaxonomyTermGroupAccessPolicyOperations() {
    $policy = AccessPolicy::create([
      'id' => 'group_policy_with_terms',
      'label' => 'Group test policy taxonomy terms',
      'target_entity_type_id' => 'taxonomy_term',
    ]);
    $policy->save();
    // Add a term to the vocabulary.
    $term = Term::create([
      'name' => 'Sometimes people are just jerks',
      'vid' => 'tags',
    ]);
    $term->save();

    $this->contentAccessPolicyManager->assign($term, $policy);

    $this->createRole([
      'access content',
      'create terms in tags',
      'delete terms in tags',
      'edit terms in tags',
      'view group_policy_with_terms taxonomy term',
      'edit group_policy_with_terms taxonomy term',
      'delete group_policy_with_terms taxonomy term',
      'set entity access policy',
      'assign group_policy_with_terms access policy',
    ], 'group_policy');

    $account = new UserSession([
      'uid' => 2,
      'roles' => ['group_policy'],
    ]);

    $result = $term->access('view', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the term.');

    $result = $term->access('update', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to edit the term.');

    $result = $term->access('delete', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to delete the term.');

    $result = $term->access('manage access', $account, TRUE);
    $this->assertTrue($result->isAllowed(), 'The user should be able to view the term Access page.');

    $this->createRole([
      'access content',
      'create terms in tags',
      'delete terms in tags',
      'edit terms in tags',
      'set entity access policy',
    ], 'authenticated');

    $account = new UserSession([
      'roles' => ['authenticated'],
    ]);

    $result = $term->access('view', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to view the term.');

    $result = $term->access('update', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not able to edit the term.');

    $result = $term->access('delete', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to delete the term.');

    $result = $term->access('manage access', $account, TRUE);
    $this->assertTrue($result->isForbidden(), 'The user should not be able to assign the policy.');

  }

}
