<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\OperatorValidationTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests EntityField boolean plugin features.
 *
 * @group access_policy
 */
class EntityFieldBooleanTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use OperatorValidationTrait;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_boolean',
      'entity_type' => 'node',
      'type' => 'boolean',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'field_boolean',
      'entity_type' => 'node',
      'bundle' => 'page',
      'label' => "boolean",
      'required' => TRUE,
      'settings' => [
        'on_label' => 'On',
        'off_label' => 'Off',
      ],
    ]);
    $field->save();
  }

  /**
   * Tests entity field standard equals.
   */
  public function testEntityFieldBoolean() {

    $policy = AccessPolicy::create([
      'id' => 'boolean',
      'label' => 'Access policy boolean',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add an equals operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_boolean');
    $access_rule->setSettings([
      'value' => 'true',
      'operator' => '=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 has foo and should be accessible.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_boolean' => TRUE,
    ]);

    $node_2 = $this->createNode([
      'title' => 'Node 2',
      'type' => 'page',
      'status' => 1,
      'field_boolean' => FALSE,
    ]);

    $this->assignAccessPolicy($node_1, 'boolean');
    $this->assignAccessPolicy($node_2, 'boolean');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'view boolean content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    // Change the access rule settings and validate again.
    $access_rule_settings = $access_rule->getSettings();
    $access_rule_settings['operator'] = '!=';
    $policy->updateAccessRule('field_boolean', $access_rule_settings);
    $policy->save();

    $this->accessPolicyValidator->resetCache();
    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_2->id(),
    ]);
  }

}
