<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\OperatorValidationTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\link\LinkItemInterface;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests EntityField boolean plugin features.
 *
 * @group access_policy
 */
class EntityFieldEmptyTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use OperatorValidationTrait;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
    'link',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);

    // The link field itself doesn't make much sense in terms of access so we
    // just check to see whether it's empty or not.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_link',
      'entity_type' => 'node',
      'type' => 'link',
      'settings' => [],
      'cardinality' => 1,
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'bundle' => 'page',
      'settings' => [
        'title' => DRUPAL_OPTIONAL,
        'link_type' => LinkItemInterface::LINK_GENERIC,
      ],
    ]);
    $field->save();
  }

  /**
   * Tests entity field standard equals.
   */
  public function testEntityFieldEmpty() {

    $policy = AccessPolicy::create([
      'id' => 'policy_basic',
      'label' => 'Basic policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add an equals operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_link');
    $access_rule->setSettings([
      'operator' => 'not empty',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 has the link and should be accessible.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_link' => 'https://foo.org/',
    ]);

    $node_2 = $this->createNode([
      'title' => 'Node 2',
      'type' => 'page',
      'status' => 1,
    ]);

    $this->assignAccessPolicy($node_1, 'policy_basic');
    $this->assignAccessPolicy($node_2, 'policy_basic');

    $admin_user = $this->createUser([
      'access content',
      'view policy_basic content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    // Change the access rule settings and validate again.
    $access_rule_settings = $access_rule->getSettings();
    $access_rule_settings['operator'] = 'empty';
    $policy->updateAccessRule('field_link', $access_rule_settings);
    $policy->save();

    $this->accessPolicyValidator->resetCache();
    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_2->id(),
    ]);
  }

}
