<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\OperatorValidationTrait;
use Drupal\node\Entity\NodeType;
use Drupal\Tests\content_moderation\Traits\ContentModerationTestTrait;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests ModerationStateTest access rule plugin.
 *
 * @group access_policy
 */
class ModerationStateTest extends AccessPolicyKernelTestBase {

  use ContentModerationTestTrait;
  use TaxonomyTestTrait;
  use OperatorValidationTrait;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'system',
    'text',
    'user',
    'datetime',
    'content_moderation',
    'workflows',
    'access_policy',
    'access_policy_test',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);
    $this->installEntitySchema('content_moderation_state');
    $this->installConfig(['content_moderation', 'system']);

    // Articles are not using content moderation.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    $nodeType = NodeType::load('page');
    $nodeType->setNewRevision(TRUE);
    $nodeType->save();

    $workflow = $this->createEditorialWorkflow();
    $workflow->getTypePlugin()->addEntityTypeAndBundle('node', 'page');
    $workflow->save();
  }

  /**
   * Tests moderation state access rule.
   */
  public function testModerationState() {
    $policy = AccessPolicy::create([
      'id' => 'archived',
      'label' => 'Archived policy',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();
    $access_rule = $this->accessRuleManager->getHandler('node', 'moderation_state');
    $access_rule->setSettings([
      'value' => [
        'editorial-archived' => 'editorial-archived',
      ],
      'operator' => 'or',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Create a basic node.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'moderation_state' => 'archived',
    ]);
    $this->assignAccessPolicy($node_1, 'archived');

    // Create an article node. This should not be excluded from results because
    // it's not covered by content moderation.
    $node_article = $this->createNode([
      'title' => 'Article',
      'type' => 'article',
      'status' => 1,
    ]);

    $user = $this->createUser([
      'access content',
      'view archived content',
      'view any archived unpublished content',
    ]);

    $user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $user, 'view');
    $this->assertTrue($access, 'The user should have access');

    $this->setCurrentUser($user);
    $this->assertQueryResults('node', [
      $node_1->id(),
      $node_article->id(),
    ]);

    // Change the access rule settings and run validation again.
    $access_rule_settings = $access_rule->getSettings();
    $access_rule_settings['operator'] = 'not';
    $policy->updateAccessRule('moderation_state', $access_rule_settings);
    $policy->save();

    $this->accessPolicyValidator->resetCache();
    $access = $this->accessPolicyValidator->validate($node_1, $user, 'view');
    $this->assertFalse($access, 'The user should not have access');
    $this->assertQueryResults('node', [
      $node_article->id(),
    ]);
  }

}
